import abc
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Generic, TypeVar

import pytest

from aioinject import Container, Object, Scoped, Transient


T = TypeVar("T")
U = TypeVar("U")
ReqT = TypeVar("ReqT")
ResT = TypeVar("ResT")


class UnusedGeneric(Generic[T]):
    def __init__(self, dependency: str) -> None:
        self.dependency = dependency


class SimpleGeneric(Generic[T]):
    def __init__(self, dependency: T) -> None:
        self.dependency = dependency


class SimpleGenericExtended(SimpleGeneric[int]):
    pass


class MultipleSimpleGeneric(Generic[T, U]):
    def __init__(self, service: T, b: U) -> None:
        self.service = service
        self.b = b


class NestedGeneric(Generic[T, U]):
    def __init__(self, simple_gen: MultipleSimpleGeneric[T, U], u: U) -> None:
        self.simple_gen = simple_gen
        self.u = u


class Something:
    def __init__(self) -> None:
        self.a = MEANING_OF_LIFE_INT


MEANING_OF_LIFE_INT = 42
MEANING_OF_LIFE_STR = "42"


@pytest.mark.parametrize(
    ("type_", "instanceof"),
    [
        (UnusedGeneric, UnusedGeneric),
        (SimpleGeneric[int], SimpleGeneric),
        (SimpleGenericExtended, SimpleGenericExtended),
    ],
)
async def test_resolve_generics(
    type_: type[object],
    instanceof: type[object],
) -> None:
    container = Container()
    container.register(Scoped(type_))
    container.register(Object(42))
    container.register(Object("42"))

    async with container.context() as ctx:
        instance = await ctx.resolve(type_)
        assert isinstance(instance, instanceof)


async def test_nested_generics() -> None:
    container = Container()
    container.register(
        Scoped(NestedGeneric[int, str]),
        Scoped(MultipleSimpleGeneric[int, str]),
        Object(MEANING_OF_LIFE_INT),
        Object(MEANING_OF_LIFE_STR),
    )

    async with container.context() as ctx:
        instance = await ctx.resolve(NestedGeneric[int, str])
        assert isinstance(instance, NestedGeneric)
        assert isinstance(instance.simple_gen, MultipleSimpleGeneric)
        assert instance.simple_gen.service == MEANING_OF_LIFE_INT
        assert instance.simple_gen.b == MEANING_OF_LIFE_STR
        assert instance.u == MEANING_OF_LIFE_STR


async def test_nested_unresolved_generic() -> None:
    @dataclass
    class NestedUnresolvedGeneric:
        service: SimpleGeneric  # type: ignore[type-arg]

    container = Container()
    obj = SimpleGeneric(MEANING_OF_LIFE_INT)
    container.register(
        Scoped(NestedUnresolvedGeneric),
        Object(obj, interface=SimpleGeneric),
    )

    async with container.context() as ctx:
        instance = await ctx.resolve(NestedUnresolvedGeneric)
        assert isinstance(instance, NestedUnresolvedGeneric)
        assert isinstance(instance.service, SimpleGeneric)
        assert instance.service.dependency == MEANING_OF_LIFE_INT


async def test_partially_resolved_generic() -> None:
    K = TypeVar("K")

    class TwoGeneric(Generic[T, K]):
        def __init__(self, a: SimpleGeneric[T], b: SimpleGeneric[K]) -> None:
            self.a = a
            self.b = b

    class UsesTwoGeneric(Generic[T]):
        def __init__(self, service: TwoGeneric[T, str]) -> None:
            self.service = service

    container = Container()
    container.register(
        Scoped(UsesTwoGeneric[int]),
        Scoped(TwoGeneric[int, str]),
        Scoped(SimpleGeneric[int]),
        Scoped(SimpleGeneric[str]),
        Object(MEANING_OF_LIFE_INT),
        Object("42"),
    )

    async with container.context() as ctx:
        instance = await ctx.resolve(UsesTwoGeneric[int])
        assert isinstance(instance, UsesTwoGeneric)
        assert isinstance(instance.service, TwoGeneric)
        assert isinstance(instance.service.a, SimpleGeneric)
        assert isinstance(instance.service.b, SimpleGeneric)
        assert instance.service.a.dependency == MEANING_OF_LIFE_INT
        assert instance.service.b.dependency == MEANING_OF_LIFE_STR


async def test_can_resolve_generic_class_without_parameters() -> None:
    class GenericClass(Generic[T]):
        def __init__(self, a: int) -> None:
            self.a = a

        def so_generic(self) -> T:  # pragma: no cover
            raise NotImplementedError

    container = Container()
    container.register(Scoped(GenericClass), Object(MEANING_OF_LIFE_INT))

    async with container.context() as ctx:
        instance = await ctx.resolve(GenericClass)
        assert isinstance(instance, GenericClass)
        assert instance.a == MEANING_OF_LIFE_INT


async def test_can_resolve_generic_iterable() -> None:
    class MiddlewareBase(abc.ABC, Generic[ReqT, ResT]):
        @abc.abstractmethod
        async def __call__(
            self,
            request: ReqT,
            handle: Callable[[ReqT], Awaitable[ResT]],
        ) -> ResT:
            pass

    class FirstMiddleware(MiddlewareBase[ReqT, ResT]):
        async def __call__(
            self,
            request: ReqT,
            handle: Callable[[ReqT], Awaitable[ResT]],
        ) -> ResT:
            return await handle(request)

    class SecondMiddleware(MiddlewareBase[ReqT, ResT]):
        async def __call__(
            self,
            request: ReqT,
            handle: Callable[[ReqT], Awaitable[ResT]],
        ) -> ResT:
            return await handle(request)

    class ThirdMiddleware(MiddlewareBase[ReqT, ResT]):
        async def __call__(
            self,
            request: ReqT,
            handle: Callable[[ReqT], Awaitable[ResT]],
        ) -> ResT:
            return await handle(request)

    container = Container()
    container.register(
        Transient(FirstMiddleware, MiddlewareBase[str, str]),
        Transient(SecondMiddleware, MiddlewareBase[int, int]),
        Transient(ThirdMiddleware, MiddlewareBase[int, int]),
    )

    async with container.context() as ctx:
        instances = await ctx.resolve(list[MiddlewareBase[int, int]])
        assert len(instances) == 2  # noqa: PLR2004
        assert isinstance(instances[0], SecondMiddleware)
        assert isinstance(instances[1], ThirdMiddleware)


async def test_can_resolve_generic_object() -> None:
    class MyGeneric(Generic[T]):
        pass

    class Dependant:
        def __init__(self, generic: MyGeneric[int]) -> None:
            self.generic = generic

    container = Container()
    container.register(Object(MyGeneric(), MyGeneric[int]))
    container.register(Scoped(Dependant))

    async with container, container.context() as ctx:
        await ctx.resolve(Dependant)



from typing import Any

import strawberry

from aioinject import Injected
from aioinject.ext.strawberry import inject
from tests.integrations.conftest import NumberService


@strawberry.type
class Query:
    @strawberry.field
    @inject
    async def number(self, service: Injected[NumberService]) -> int:
        return service.get_number()

    @strawberry.field
    @inject
    async def number_with_info(
        self, service: Injected[NumberService], info: strawberry.Info[Any, Any]
    ) -> int:
        assert info
        assert isinstance(info, strawberry.Info)
        return service.get_number()



from __future__ import annotations

import inspect
import typing
from collections.abc import Callable, Iterator
from typing import Any

import strawberry
from strawberry.extensions import SchemaExtension


__all__ = ["AioInjectExtension", "inject"]

from strawberry.utils.typing import is_generic_alias

from aioinject import Container, Context, SyncContainer, SyncContext
from aioinject._types import P, T, safe_issubclass
from aioinject.decorators import ContextParameter, base_inject


def _find_strawberry_info_parameter(
    function: Callable[..., Any],
) -> inspect.Parameter | None:
    signature = inspect.signature(function)
    for p in signature.parameters.values():
        annotation = p.annotation

        if is_generic_alias(annotation):
            annotation = typing.get_origin(annotation)

        try:
            if safe_issubclass(annotation, strawberry.Info):
                return p
        except TypeError:  # pragma: no cover
            continue
    return None


def _default_context_getter(context: Any) -> Context | SyncContext:
    return context["aioinject_context"]


def _default_context_setter(
    context: Any, aioinject_context: Context | SyncContext
) -> None:
    context["aioinject_context"] = aioinject_context


def inject(
    function: Callable[P, T],
    context_getter: Callable[
        [Any], Context | SyncContext
    ] = _default_context_getter,
) -> Callable[P, T]:
    info_parameter = _find_strawberry_info_parameter(function)
    info_parameter_name = (
        info_parameter.name if info_parameter else "aioinject_info"
    )

    return base_inject(
        function=function,
        context_parameters=(
            ContextParameter(
                name=info_parameter_name,
                type_=strawberry.Info,
                remove=info_parameter is None,
            ),
        ),
        context_getter=lambda args, kwargs: context_getter(  # noqa: ARG005
            kwargs[info_parameter_name].context
        ),
        enter_context=True,
    )


class AioInjectExtension(SchemaExtension):
    def __init__(
        self,
        container: Container | SyncContainer,
        context_setter: Callable[
            [Any, Context | SyncContext], None
        ] = _default_context_setter,
    ) -> None:
        self.container = container
        self._context_setter = context_setter

    def on_operation(
        self,
    ) -> Iterator[None]:
        self._context_setter(
            self.execution_context.context, self.container.root
        )
        yield



from collections.abc import Awaitable, Callable
from typing import Any

import pytest
from aiogram import BaseMiddleware, Router
from aiogram.types import TelegramObject

import aioinject
from aioinject import INJECTED, Container, Injected, Object
from aioinject.ext.aiogram import AioInjectMiddleware, inject


_NUMBER = 42


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(Object(_NUMBER))
    return container


async def test_handler(container: Container) -> None:
    middleware = AioInjectMiddleware(container=container)
    event_ = object()
    data_: dict[str, Any] = {}

    async def handler(
        event: object,
        data: dict[str, Any],
    ) -> None:
        assert event is event_
        assert data is data_
        assert isinstance(data["aioinject_context"], aioinject.Context)

    await middleware(handler=handler, event=event_, data=data_)  # type: ignore[arg-type]


async def test_can_inject_into_middleware(container: Container) -> None:
    class CustomMiddleware(BaseMiddleware):
        @inject
        async def __call__(
            self,
            handler: Callable[  # noqa: ARG002
                [TelegramObject, dict[str, Any]], Awaitable[Any]
            ],
            event: TelegramObject,  # noqa: ARG002
            data: dict[str, Any],  # noqa: ARG002
            number: Injected[int] = INJECTED,
        ) -> Any:
            assert number is _NUMBER

    async def call_next(event: object, data: dict[str, Any]) -> None:
        await CustomMiddleware()(None, event, data)  # type: ignore[arg-type]

    middleware = AioInjectMiddleware(container=container)
    await middleware(handler=call_next, event=object(), data={})  # type: ignore[arg-type]


async def test_no_context_err() -> None:
    @inject
    async def handler(
        event: object,
        data: dict[str, Any],
    ) -> None:
        pass

    with pytest.raises(ValueError) as exc_info:  # noqa: PT011
        await handler(object(), {})

    assert (
        str(exc_info.value)
        == "Could not find aioinject context, did you forget to add AioinjectMiddleware?"
    )


def test_add_to_router() -> None:
    middleware = AioInjectMiddleware(container=Container())

    router = Router()
    middleware.add_to_router(router=router)

    for observer in router.observers.values():
        assert observer.outer_middleware[0] is middleware



from __future__ import annotations

import collections
from collections.abc import Sequence
from typing import TYPE_CHECKING

from aioinject.validation.errors import RuleViolation, ValidationError


if TYPE_CHECKING:
    from aioinject import Container, SyncContainer
    from aioinject.validation.abc import ValidationRule


def validate_or_err(
    container: SyncContainer | Container,
    rules: Sequence[ValidationRule],
) -> None:
    errors: list[RuleViolation] = []
    for rule in rules:
        errors.extend(rule.validate(container))

    if not errors:
        return

    parts = [""]
    violations_by_group = collections.defaultdict(list)
    for violation in errors:
        violations_by_group[violation.code].append(violation)

    for code, violations in violations_by_group.items():
        parts.append(f"  {code}:")
        parts.extend(f"    {violation.message}" for violation in violations)

    error_message = "\n".join(parts)
    raise ValidationError(error_message)



import contextlib
import dataclasses
import statistics
import time
from collections.abc import Awaitable, Callable, Iterator, Sequence
from typing import Final


BenchmarkedFunction = Callable[["BenchmarkContext"], Awaitable[None]]


@dataclasses.dataclass
class ProjectUrl:
    url: str


Extra = ProjectUrl


class BenchmarkEntry:
    def __init__(
        self,
        func: BenchmarkedFunction,
        name: str,
        max_iterations: int | None,
        extras: Sequence[Extra],
    ) -> None:
        self.func = func
        self.name = name
        self.max_iterations = max_iterations
        self.extras = extras

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"


def bench(
    name: str,
    max_iterations: int | None = None,
    extras: Sequence[Extra] = (),
) -> Callable[[BenchmarkedFunction], BenchmarkEntry]:
    def inner(func: BenchmarkedFunction) -> BenchmarkEntry:
        return BenchmarkEntry(
            func, name=name, max_iterations=max_iterations, extras=extras
        )

    return inner


class BenchmarkContext:
    def __init__(self, rounds: int) -> None:
        self.durations: list[float] = []
        self.rounds: Final = rounds

    @contextlib.contextmanager
    def round(self) -> Iterator[None]:
        start = time.perf_counter()
        yield
        elapsed = time.perf_counter() - start
        self.durations.append(elapsed)


@dataclasses.dataclass
class BenchmarkParameters:
    rounds: int


@dataclasses.dataclass
class BenchmarkResult:
    benchmark: BenchmarkEntry
    params: BenchmarkParameters

    name: str
    rounds: int
    mean: float
    median: float
    total: float

    extrapolated: bool


class BenchmarkCollection:
    def __init__(self, benchmarks: Sequence[BenchmarkEntry] = ()) -> None:
        self.benchmarks = list(benchmarks)

    def bench(
        self,
        name: str,
        max_iterations: int | None = None,
        extras: Sequence[Extra] = (),
    ) -> Callable[[BenchmarkedFunction], BenchmarkEntry]:
        def inner(func: BenchmarkedFunction) -> BenchmarkEntry:
            entry = BenchmarkEntry(
                func, name=name, max_iterations=max_iterations, extras=extras
            )
            self.benchmarks.append(entry)
            return entry

        return inner


class Benchmark:
    def __init__(
        self,
        benchmarks: list[BenchmarkEntry],
    ) -> None:
        self.benchmarks = benchmarks

    async def run(
        self,
        rounds: Sequence[int],
    ) -> Sequence[BenchmarkResult]:
        results = []
        for benchmark in self.benchmarks:
            for iterations in rounds:
                benchmark_results = [
                    await self.run_benchmark(benchmark, iterations)
                    for _ in range(5)
                ]
                best_result = min(benchmark_results, key=lambda r: r.total)
                results.append(best_result)

        return results

    async def run_benchmark(
        self, benchmark: BenchmarkEntry, rounds: int
    ) -> BenchmarkResult:
        extrapolated = (
            benchmark.max_iterations < rounds
            if benchmark.max_iterations
            else False
        )
        actual_rounds = (
            min(benchmark.max_iterations, rounds)
            if extrapolated and benchmark.max_iterations
            else rounds
        )
        context = BenchmarkContext(rounds=actual_rounds)
        await benchmark.func(context)
        context.durations.sort()
        result = BenchmarkResult(
            params=BenchmarkParameters(rounds=rounds),
            name=benchmark.name,
            rounds=actual_rounds,
            mean=statistics.mean(context.durations),
            median=statistics.median(context.durations),
            total=sum(context.durations),
            extrapolated=extrapolated,
            benchmark=benchmark,
        )
        assert len(context.durations) == actual_rounds  # noqa: S101
        return result



from collections.abc import Sequence
from typing import Protocol

from aioinject import Container, SyncContainer
from aioinject.validation.errors import RuleViolation


class ValidationRule(Protocol):
    def validate(
        self, container: Container | SyncContainer
    ) -> Sequence[RuleViolation]: ...



from typing import Any

import aioinject
from aioinject import Context
from aioinject.context import ProviderRecord
from aioinject.extensions import OnResolveExtension


class _TestExtension(OnResolveExtension):
    async def on_resolve(
        self,
        context: Context,
        provider: ProviderRecord[Any],
        instance: Any,
    ) -> None: ...


async def _pass() -> None:
    aioinject.Container(extensions=[_TestExtension()])



from typing import Any

import strawberry
import uvicorn
from strawberry import Schema
from strawberry.asgi import GraphQL

import aioinject
from aioinject import Injected
from aioinject.ext.strawberry import AioInjectExtension, inject  # (1)!


container = aioinject.Container()
container.register(aioinject.Object(42))


@strawberry.type
class Query:
    @strawberry.field
    @inject
    async def number(self, number: Injected[int]) -> int:
        return number


def create_app() -> GraphQL[Any, Any]:
    schema = Schema(
        query=Query,
        extensions=[
            AioInjectExtension(container=container),
        ],
    )
    return GraphQL(schema=schema)


if __name__ == "__main__":
    uvicorn.run("main:create_app", factory=True, reload=True)



import aioinject


container = aioinject.SyncContainer()
container.register(aioinject.Transient(list))

with container.context() as ctx:
    object_1 = ctx.resolve(list)
    object_2 = ctx.resolve(list)
    assert object_1 is not object_2



from aioinject.extensions._abc import (
    Extension,
    LifespanExtension,
    LifespanSyncExtension,
    OnInitExtension,
    OnResolveContextExtension,
    OnResolveExtension,
    OnResolveSyncExtension,
    ProviderExtension,
)


__all__ = [
    "Extension",
    "LifespanExtension",
    "LifespanSyncExtension",
    "OnInitExtension",
    "OnResolveContextExtension",
    "OnResolveExtension",
    "OnResolveSyncExtension",
    "ProviderExtension",
]



from __future__ import annotations

from aioinject import Container, Singleton


cont = Container()


# notice that C is not defined yet
def get_c() -> C:
    return C()


cont.register(Singleton(get_c))


class C: ...
